/* SD Card FAT File System Demo
 * http://frank.circleofcurrent.com/musicclock/
 * Copyright (c) 2011 Frank Zhao
 * 
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of either the GNU General Public License version 2
 * or the GNU Lesser General Public License version 2.1, both as
 * published by the Free Software Foundation.
 *
 */

// please see http://www.nongnu.org/avr-libc/user-manual/modules.html for info on these modules

#include <avr/io.h> // allows access to AVR built-in peripherals and registers
#include <avr/pgmspace.h> // allows access to flash memory
#include <util/delay.h> // provides delays
#include <stdio.h> // allows formatted debug output
#include <avr/power.h> // used for setting clock prescaler
#include "libs/ff.h" // use FatFs from Elm-Chan.org
#include "libs/diskio.h" // use FatFs from Elm-Chan.org (diskio used for disk initialization)
#include "libs/usb_serial.h" // provides USB-to-serial functionality

// this is a wrapper for usb_putchar that allows me to use printf from stdio.h
static inline int debug_putchar(char c, FILE* stream)
{
    return usb_serial_putchar(c); // refer to http://www.pjrc.com/teensy/usb_serial.html
}

// setup stream for printf from stdio.h
static FILE mystdout = FDEV_SETUP_STREAM(debug_putchar, NULL, _FDEV_SETUP_WRITE);

// program entry point
int main()
{
    clock_prescale_set(clock_div_2); // this is because we are running at 3.3V
    // see http://www.nongnu.org/avr-libc/user-manual/group__avr__power.html and http://www.pjrc.com/teensy/3volt.html
    
    stdout = &mystdout; // setup stream for printf from stdio.h
    
    // initialize stuff
    usb_init();
    
    while (usb_configured() == 0) _delay_ms(5); // waits until USB is ready
    
    _delay_ms(5000); // 5 seconds for you to connect hyperterminal
    
    // here's the start of the demonstration
    printf_P(PSTR("SD Card FAT File System Demo\r\n"));
    
    FATFS filesystem;
    FRESULT errCode;
    
    printf_P(PSTR("Mounting File System\r\n"));
    errCode = f_mount(0, &filesystem);
    if (errCode != FR_OK)
    {
        printf_P(PSTR("Error: f_mount failed, error code: %d\r\n"), errCode);
        while (1); // freeze
    }
    
    printf_P(PSTR("Opening Root Directory\r\n"));
    DIR rootDir;
    errCode = f_opendir(&rootDir, "/");
    if (errCode != FR_OK)
    {
        printf_P(PSTR("Error: f_opendir failed, error code: %d\r\n"), errCode);
        while (1); // freeze
    }
    
    // read all file entries
    while (1)
    {
        FILINFO fileEntry;
        errCode = f_readdir(&rootDir, &fileEntry);
        if (errCode != FR_OK)
        {
            printf_P(PSTR("Error: f_readdir failed, error code: %d\r\n"), errCode);
            while (1); // freeze
        }
        
        if (fileEntry.fname[0] == 0)
        {
            printf_P(PSTR("End of All Files\r\n"));
            break;
        }
        else if (fileEntry.fname[0] == '.')
        {
            printf_P(PSTR("Skipping \"%s\"\r\n\r\n"), fileEntry.fname);
        }
        else if (fileEntry.fattrib & AM_DIR)
        {
            printf_P(PSTR("Directory Entry: %s\r\n\r\n"), fileEntry.fname);
        }
        else
        {
            printf_P(PSTR("File Entry: %s\r\n"), fileEntry.fname);
            
            // open file
            FIL file;
            errCode = f_open(&file, fileEntry.fname, FA_READ);
            if (errCode != FR_OK)
            {
                printf_P(PSTR("Error: f_open failed, error code: %d\r\n"), errCode);
                while (1); // freeze
            }
            
            // read file until end
            while (1)
            {
                uint8_t buff[64];
                UINT read;
                errCode = f_read(&file, buff, 64, &read);
                if (errCode != FR_OK)
                {
                    printf_P(PSTR("Error: f_close failed, error code: %d\r\n"), errCode);
                    while (1); // freeze
                }
                
                for (UINT i = 0; i < read; i++)
                {
                    debug_putchar(buff[i], 0);
                }
                
                if (read != 64)
                {
                    printf_P(PSTR("\r\nEnd of File\r\n\r\n"));
                    break;
                }
            }
            
            errCode = f_close(&file);
            if (errCode != FR_OK)
            {
                printf_P(PSTR("Error: f_close failed, error code: %d\r\n"), errCode);
            }
        }
    }
    
    while (1); // infinite loop
    return 0;
}